@@ -20,24 +20,25 @@ import static org.apache.qpid.jms.provider.amqp.AmqpSupport.MODIFIED_FAILED;
 import static org.apache.qpid.jms.provider.amqp.AmqpSupport.MODIFIED_FAILED_UNDELIVERABLE;
 import static org.apache.qpid.jms.provider.amqp.AmqpSupport.REJECTED;
 
-import java.io.IOException;
 import java.util.ArrayList;
 import java.util.ListIterator;
 import java.util.concurrent.ScheduledFuture;
 
+import javax.jms.Session;
+
 import org.apache.qpid.jms.JmsDestination;
-import org.apache.qpid.jms.JmsOperationTimedOutException;
-import org.apache.qpid.jms.exceptions.JmsExceptionSupport;
 import org.apache.qpid.jms.message.JmsInboundMessageDispatch;
 import org.apache.qpid.jms.message.JmsMessage;
 import org.apache.qpid.jms.meta.JmsConsumerId;
 import org.apache.qpid.jms.meta.JmsConsumerInfo;
 import org.apache.qpid.jms.provider.AsyncResult;
 import org.apache.qpid.jms.provider.ProviderConstants.ACK_TYPE;
+import org.apache.qpid.jms.provider.ProviderException;
 import org.apache.qpid.jms.provider.ProviderListener;
 import org.apache.qpid.jms.provider.WrappedAsyncResult;
 import org.apache.qpid.jms.provider.amqp.message.AmqpCodec;
-import org.apache.qpid.jms.util.IOExceptionSupport;
+import org.apache.qpid.jms.provider.exceptions.ProviderExceptionSupport;
+import org.apache.qpid.jms.provider.exceptions.ProviderOperationTimedOutException;
 import org.apache.qpid.proton.amqp.Binary;
 import org.apache.qpid.proton.amqp.messaging.Accepted;
 import org.apache.qpid.proton.amqp.messaging.Released;
@@ -55,20 +56,25 @@ public class AmqpConsumer extends AmqpAbstractResource<JmsConsumerInfo, Receiver
     private static final Logger LOG = LoggerFactory.getLogger(AmqpConsumer.class);
 
     protected final AmqpSession session;
+    protected final int acknowledgementMode;
     protected AsyncResult stopRequest;
     protected AsyncResult pullRequest;
     protected long incomingSequence;
-    protected long deliveredCount;
+    protected int deliveredCount;
+    protected int dispatchedCount;
     protected boolean deferredClose;
 
     public AmqpConsumer(AmqpSession session, JmsConsumerInfo info, Receiver receiver) {
         super(info, receiver, session);
 
         this.session = session;
+        this.acknowledgementMode = info.getAcknowledgementMode();
     }
 
     @Override
     public void close(AsyncResult request) {
+        acknowledgeUndeliveredRecoveredMessages();
+
         // If we have pending deliveries we remain open to allow for ACK or for a
         // pending transaction that this consumer is active in to complete.
         if (shouldDeferClose()) {
@@ -79,6 +85,27 @@ public class AmqpConsumer extends AmqpAbstractResource<JmsConsumerInfo, Receiver
         }
     }
 
+    private void acknowledgeUndeliveredRecoveredMessages() {
+        if(acknowledgementMode == Session.CLIENT_ACKNOWLEDGE) {
+            // Send dispositions for any messages which were previously delivered and
+            // session recovered, but were then not delivered again afterwards.
+            Delivery delivery = getEndpoint().head();
+            while (delivery != null) {
+                Delivery current = delivery;
+                delivery = delivery.next();
+
+                if (!(current.getContext() instanceof JmsInboundMessageDispatch)) {
+                    continue;
+                }
+
+                JmsInboundMessageDispatch envelope = (JmsInboundMessageDispatch) current.getContext();
+                if (envelope.isRecovered() && !envelope.isDelivered()) {
+                    handleDisposition(envelope, current, MODIFIED_FAILED);
+                }
+            }
+        }
+    }
+
     /**
      * Starts the consumer by setting the link credit to the given prefetch value.
      *
@@ -109,7 +136,29 @@ public class AmqpConsumer extends AmqpAbstractResource<JmsConsumerInfo, Receiver
                 request.onSuccess();
             } else {
                 // There are still deliveries to process, wait for them to be.
-                stopRequest = request;
+                if (getDrainTimeout() > 0) {
+                    // If the remote doesn't respond we will close the consumer and break any
+                    // blocked receive or stop calls that are waiting, unless the consumer is
+                    // a participant in a transaction in which case we will just fail the request
+                    // and leave the consumer open since the TX needs it to remain active.
+                    final ScheduledFuture<?> future = getSession().schedule(() -> {
+                        LOG.trace("Consumer {} stop timed out awaiting message processing", getConsumerId());
+                        ProviderException cause = new ProviderOperationTimedOutException("Consumer stop timed out awaiting message processing");
+                        if (session.isTransacted() && session.getTransactionContext().isInTransaction(getConsumerId())) {
+                            stopRequest.onFailure(cause);
+                            stopRequest = null;
+                        } else {
+                            closeResource(session.getProvider(), cause, false);
+                            session.getProvider().pumpToProtonTransport();
+                        }
+                    }, getDrainTimeout());
+
+                    stopRequest = new ScheduledRequest(future, request);
+                } else {
+                    stopRequest = request;
+                }
+
+                LOG.trace("Consumer {} stop awaiting queued delivery processing", getConsumerId());
             }
         } else {
             // TODO: We don't actually want the additional messages that could be sent while
@@ -125,18 +174,15 @@ public class AmqpConsumer extends AmqpAbstractResource<JmsConsumerInfo, Receiver
                 // blocked receive or stop calls that are waiting, unless the consumer is
                 // a participant in a transaction in which case we will just fail the request
                 // and leave the consumer open since the TX needs it to remain active.
-                final ScheduledFuture<?> future = getSession().schedule(new Runnable() {
-                    @Override
-                    public void run() {
-                        LOG.trace("Consumer {} drain request timed out", getConsumerId());
-                        Exception cause = new JmsOperationTimedOutException("Remote did not respond to a drain request in time");
-                        if (session.isTransacted() && session.getTransactionContext().isInTransaction(getConsumerId())) {
-                            stopRequest.onFailure(cause);
-                            stopRequest = null;
-                        } else {
-                            closeResource(session.getProvider(), cause, false);
-                            session.getProvider().pumpToProtonTransport();
-                        }
+                final ScheduledFuture<?> future = getSession().schedule(() -> {
+                    LOG.trace("Consumer {} drain request timed out", getConsumerId());
+                    ProviderException cause = new ProviderOperationTimedOutException("Remote did not respond to a drain request in time");
+                    if (session.isTransacted() && session.getTransactionContext().isInTransaction(getConsumerId())) {
+                        stopRequest.onFailure(cause);
+                        stopRequest = null;
+                    } else {
+                        closeResource(session.getProvider(), cause, false);
+                        session.getProvider().pumpToProtonTransport();
                     }
                 }, getDrainTimeout());
 
@@ -148,22 +194,17 @@ public class AmqpConsumer extends AmqpAbstractResource<JmsConsumerInfo, Receiver
     private void stopOnSchedule(long timeout, final AsyncResult request) {
         LOG.trace("Consumer {} scheduling stop", getConsumerId());
         // We need to drain the credit if no message(s) arrive to use it.
-        final ScheduledFuture<?> future = getSession().schedule(new Runnable() {
-            @Override
-            public void run() {
-                LOG.trace("Consumer {} running scheduled stop", getConsumerId());
-                if (getEndpoint().getRemoteCredit() != 0) {
-                    stop(request);
-                    session.getProvider().pumpToProtonTransport(request);
-                }
-            }
+        final ScheduledFuture<?> future = getSession().schedule(() -> {
+            LOG.trace("Consumer {} running scheduled stop", getConsumerId());
+            stop(request);
+            session.getProvider().pumpToProtonTransport(request);
         }, timeout);
 
         stopRequest = new ScheduledRequest(future, request);
     }
 
     @Override
-    public void processFlowUpdates(AmqpProvider provider) throws IOException {
+    public void processFlowUpdates(AmqpProvider provider) throws ProviderException {
         // Check if we tried to stop and have now run out of credit, and
         // processed all locally queued messages
         if (stopRequest != null) {
@@ -206,34 +247,37 @@ public class AmqpConsumer extends AmqpAbstractResource<JmsConsumerInfo, Receiver
             delivery = delivery.next();
 
             if (!(current.getContext() instanceof JmsInboundMessageDispatch)) {
-                LOG.debug("{} Found incomplete delivery with no context during recover processing", AmqpConsumer.this);
+                LOG.debug("{} Found incomplete delivery with no context during session acknowledge processing", AmqpConsumer.this);
                 continue;
             }
 
             JmsInboundMessageDispatch envelope = (JmsInboundMessageDispatch) current.getContext();
-            if (envelope.isDelivered()) {
+            if(ackType == ACK_TYPE.SESSION_SHUTDOWN && (envelope.isDelivered() || envelope.isRecovered())) {
+                handleDisposition(envelope, current, MODIFIED_FAILED);
+            } else if (envelope.isDelivered()) {
+                final DeliveryState disposition;
+
                 switch (ackType) {
                     case ACCEPTED:
-                        current.disposition(Accepted.getInstance());
+                        disposition = Accepted.getInstance();
                         break;
                     case RELEASED:
-                        current.disposition(Released.getInstance());
+                        disposition = Released.getInstance();
                         break;
                     case REJECTED:
-                        current.disposition(REJECTED);
+                        disposition = REJECTED;
                         break;
                     case MODIFIED_FAILED:
-                        current.disposition(MODIFIED_FAILED);
+                        disposition = MODIFIED_FAILED;
                         break;
                     case MODIFIED_FAILED_UNDELIVERABLE:
-                        current.disposition(MODIFIED_FAILED_UNDELIVERABLE);
+                        disposition = MODIFIED_FAILED_UNDELIVERABLE;
                         break;
                     default:
                         throw new IllegalArgumentException("Invalid acknowledgement type specified: " + ackType);
                 }
 
-                current.settle();
-                deliveredCount--;
+                handleDisposition(envelope, current, disposition);
             }
         }
 
@@ -258,60 +302,79 @@ public class AmqpConsumer extends AmqpAbstractResource<JmsConsumerInfo, Receiver
             return;
         }
 
-        if (ackType.equals(ACK_TYPE.DELIVERED)) {
-            LOG.debug("Delivered Ack of message: {}", envelope);
-            deliveredCount++;
-            envelope.setDelivered(true);
-            delivery.setDefaultDeliveryState(MODIFIED_FAILED);
-            sendFlowIfNeeded();
-            return;
-        } else if (ackType.equals(ACK_TYPE.ACCEPTED)) {
-            // A Consumer may not always send a DELIVERED ack so we need to
-            // check to ensure we don't add too much credit to the link.
-            if (!envelope.isDelivered()) {
-                sendFlowIfNeeded();
-            }
-            LOG.debug("Accepted Ack of message: {}", envelope);
-            if (!delivery.remotelySettled()) {
-                if (session.isTransacted() && !getResourceInfo().isBrowser()) {
+        switch (ackType) {
+            case DELIVERED:
+                handleDelivered(envelope, delivery);
+                break;
+            case ACCEPTED:
+                handleAccepted(envelope, delivery);
+                break;
+            case REJECTED:
+                handleDisposition(envelope, delivery, REJECTED);
+                break;
+            case RELEASED:
+                handleDisposition(envelope, delivery, Released.getInstance());
+                break;
+            case MODIFIED_FAILED:
+                handleDisposition(envelope, delivery, MODIFIED_FAILED);
+                break;
+            case MODIFIED_FAILED_UNDELIVERABLE:
+                handleDisposition(envelope, delivery, MODIFIED_FAILED_UNDELIVERABLE);
+                break;
+            default:
+                LOG.warn("Unsupported Ack Type for message: {}", envelope);
+                throw new IllegalArgumentException("Unknown Acknowledgement type");
+        }
 
-                    if (session.isTransactionFailed()) {
-                        LOG.trace("Skipping ack of message {} in failed transaction.", envelope);
-                        return;
-                    }
+        sendFlowIfNeeded();
+        tryCompleteDeferredClose();
+    }
 
-                    Binary txnId = session.getTransactionContext().getAmqpTransactionId();
-                    if (txnId != null) {
-                        delivery.disposition(session.getTransactionContext().getTxnAcceptState());
-                        delivery.settle();
-                        session.getTransactionContext().registerTxConsumer(this);
-                    }
-                } else {
-                    delivery.disposition(Accepted.getInstance());
+    private void handleDelivered(JmsInboundMessageDispatch envelope, Delivery delivery) {
+        LOG.debug("Delivered Ack of message: {}", envelope);
+        deliveredCount++;
+        envelope.setRecovered(false);
+        envelope.setDelivered(true);
+        delivery.setDefaultDeliveryState(MODIFIED_FAILED);
+    }
+
+    private void handleAccepted(JmsInboundMessageDispatch envelope, Delivery delivery) {
+        LOG.debug("Accepted Ack of message: {}", envelope);
+        if (!delivery.remotelySettled()) {
+            if (session.isTransacted() && !getResourceInfo().isBrowser()) {
+
+                if (session.isTransactionFailed()) {
+                    LOG.trace("Skipping ack of message {} in failed transaction.", envelope);
+                    return;
+                }
+
+                Binary txnId = session.getTransactionContext().getAmqpTransactionId();
+                if (txnId != null) {
+                    delivery.disposition(session.getTransactionContext().getTxnAcceptState());
                     delivery.settle();
+                    session.getTransactionContext().registerTxConsumer(this);
                 }
             } else {
+                delivery.disposition(Accepted.getInstance());
                 delivery.settle();
             }
-        } else if (ackType.equals(ACK_TYPE.MODIFIED_FAILED)) {
-            settleDelivery(delivery, MODIFIED_FAILED);
-        } else if (ackType.equals(ACK_TYPE.MODIFIED_FAILED_UNDELIVERABLE)) {
-            settleDelivery(delivery, MODIFIED_FAILED_UNDELIVERABLE);
-        } else if (ackType.equals(ACK_TYPE.REJECTED)) {
-            settleDelivery(delivery, REJECTED);
-        } else if (ackType.equals(ACK_TYPE.RELEASED)) {
-            delivery.disposition(Released.getInstance());
-            delivery.settle();
         } else {
-            LOG.warn("Unsupported Ack Type for message: {}", envelope);
-            return;
+            delivery.settle();
         }
 
         if (envelope.isDelivered()) {
             deliveredCount--;
         }
+        dispatchedCount--;
+    }
 
-        tryCompleteDeferredClose();
+    private void handleDisposition(JmsInboundMessageDispatch envelope, Delivery delivery, DeliveryState outcome) {
+        delivery.disposition(outcome);
+        delivery.settle();
+        if (envelope.isDelivered()) {
+            deliveredCount--;
+        }
+        dispatchedCount--;
     }
 
     /**
@@ -331,11 +394,10 @@ public class AmqpConsumer extends AmqpAbstractResource<JmsConsumerInfo, Receiver
 
         int currentCredit = getEndpoint().getCredit();
         if (currentCredit <= prefetchSize * 0.5) {
-            int prefetchedMessageCount = getResourceInfo().getPrefetchedMessageCount();
+            int potentialPrefetch = currentCredit + (dispatchedCount - deliveredCount);
 
-            int potentialPrefetch = currentCredit + prefetchedMessageCount;
             if (potentialPrefetch <= prefetchSize * 0.7) {
-                int additionalCredit = prefetchSize - currentCredit - prefetchedMessageCount;
+                int additionalCredit = prefetchSize - potentialPrefetch;
 
                 LOG.trace("Consumer {} granting additional credit: {}", getConsumerId(), additionalCredit);
                 getEndpoint().flow(additionalCredit);
@@ -378,11 +440,20 @@ public class AmqpConsumer extends AmqpAbstractResource<JmsConsumerInfo, Receiver
                     envelope.getMessage().getFacade().getRedeliveryCount() + 1);
                 envelope.setEnqueueFirst(true);
                 envelope.setDelivered(false);
+                if(acknowledgementMode == Session.CLIENT_ACKNOWLEDGE) {
+                    envelope.setRecovered(true);
+                }
 
                 redispatchList.add(envelope);
             }
         }
 
+        // Previously delivered messages should be tagged as dispatched messages again so we
+        // can properly compute the next credit refresh, so subtract them from both the delivered
+        // and dispatched counts and then dispatch them again as a new message.
+        deliveredCount -= redispatchList.size();
+        dispatchedCount -= redispatchList.size();
+
         ListIterator<JmsInboundMessageDispatch> reverseIterator = redispatchList.listIterator(redispatchList.size());
         while (reverseIterator.hasPrevious()) {
             deliver(reverseIterator.previous());
@@ -443,7 +514,11 @@ public class AmqpConsumer extends AmqpAbstractResource<JmsConsumerInfo, Receiver
     }
 
     @Override
-    public void processDeliveryUpdates(AmqpProvider provider, Delivery delivery) throws IOException {
+    public void processDeliveryUpdates(AmqpProvider provider, Delivery delivery) throws ProviderException {
+        if(delivery.getDefaultDeliveryState() == null){
+            delivery.setDefaultDeliveryState(Released.getInstance());
+        }
+
         if (delivery.isReadable() && !delivery.isPartial()) {
             LOG.trace("{} has incoming Message(s).", this);
             try {
@@ -456,7 +531,7 @@ public class AmqpConsumer extends AmqpAbstractResource<JmsConsumerInfo, Receiver
                     }
                 }
             } catch (Exception e) {
-                throw IOExceptionSupport.create(e);
+                throw ProviderExceptionSupport.createNonFatalOrPassthrough(e);
             }
         }
 
@@ -475,8 +550,6 @@ public class AmqpConsumer extends AmqpAbstractResource<JmsConsumerInfo, Receiver
     }
 
     private boolean processDelivery(Delivery incoming) throws Exception {
-        incoming.setDefaultDeliveryState(Released.getInstance());
-
         JmsMessage message = null;
         try {
             message = AmqpCodec.decodeMessage(this, getEndpoint().recv()).asJmsMessage();
@@ -487,7 +560,11 @@ public class AmqpConsumer extends AmqpAbstractResource<JmsConsumerInfo, Receiver
             //        In the future once the JMS mapping is complete we should be
             //        able to convert everything to some message even if its just
             //        a bytes messages as a fall back.
-            settleDelivery(incoming, MODIFIED_FAILED_UNDELIVERABLE);
+            incoming.disposition(MODIFIED_FAILED_UNDELIVERABLE);
+            incoming.settle();
+            // TODO: this flows credit, which we might not want, e.g if
+            // a drain was issued to stop the link.
+            sendFlowIfNeeded();
             return false;
         }
 
@@ -500,6 +577,7 @@ public class AmqpConsumer extends AmqpAbstractResource<JmsConsumerInfo, Receiver
             JmsInboundMessageDispatch envelope = new JmsInboundMessageDispatch(getNextIncomingSequenceNumber());
             envelope.setMessage(message);
             envelope.setConsumerId(getResourceInfo().getId());
+            envelope.setConsumerInfo(getResourceInfo());
             // Store link to delivery in the hint for use in acknowledge requests.
             envelope.setProviderHint(incoming);
             envelope.setMessageId(message.getFacade().getProviderMessageIdObject());
@@ -557,19 +635,12 @@ public class AmqpConsumer extends AmqpAbstractResource<JmsConsumerInfo, Receiver
         return "AmqpConsumer { " + getResourceInfo().getId() + " }";
     }
 
-    protected void settleDelivery(Delivery incoming, DeliveryState state) {
-        incoming.disposition(state);
-        incoming.settle();
-        // TODO: this flows credit, which we might not want, e.g if
-        // a drain was issued to stop the link.
-        sendFlowIfNeeded();
-    }
-
     protected void deliver(JmsInboundMessageDispatch envelope) throws Exception {
         if (!deferredClose) {
             ProviderListener listener = session.getProvider().getProviderListener();
             if (listener != null) {
                 LOG.debug("Dispatching received message: {}", envelope);
+                dispatchedCount++;
                 listener.onInboundMessage(envelope);
             } else {
                 LOG.error("Provider listener is not set, message will be dropped: {}", envelope);
@@ -593,7 +664,7 @@ public class AmqpConsumer extends AmqpAbstractResource<JmsConsumerInfo, Receiver
     }
 
     @Override
-    public void handleResourceClosure(AmqpProvider provider, Throwable cause) {
+    public void handleResourceClosure(AmqpProvider provider, ProviderException cause) {
         AmqpConnection connection = session.getConnection();
         AmqpSubscriptionTracker subTracker = connection.getSubTracker();
         JmsConsumerInfo consumerInfo = getResourceInfo();
@@ -646,15 +717,13 @@ public class AmqpConsumer extends AmqpAbstractResource<JmsConsumerInfo, Receiver
             Delivery current = delivery;
             delivery = delivery.next();
 
-            if (!(current.getContext() instanceof JmsInboundMessageDispatch)) {
+            if (current.getContext() instanceof JmsInboundMessageDispatch) {
+                JmsInboundMessageDispatch envelope = (JmsInboundMessageDispatch) current.getContext();
+                if (!envelope.isDelivered()) {
+                    handleDisposition(envelope, current, Released.getInstance());
+                }
+            } else {
                 LOG.debug("{} Found incomplete delivery with no context during release processing", AmqpConsumer.this);
-                continue;
-            }
-
-            JmsInboundMessageDispatch envelope = (JmsInboundMessageDispatch) current.getContext();
-            if (!envelope.isDelivered()) {
-                current.disposition(Released.getInstance());
-                current.settle();
             }
         }
     }
@@ -681,9 +750,9 @@ public class AmqpConsumer extends AmqpAbstractResource<JmsConsumerInfo, Receiver
     private final class DeferredCloseRequest implements AsyncResult {
 
         @Override
-        public void onFailure(Throwable result) {
+        public void onFailure(ProviderException result) {
             LOG.trace("Failed deferred close of consumer: {} - {}", getConsumerId(), result.getMessage());
-            getParent().getProvider().fireNonFatalProviderException(JmsExceptionSupport.create(result));
+            getParent().getProvider().fireNonFatalProviderException(ProviderExceptionSupport.createNonFatalOrPassthrough(result));
         }
 
         @Override
@@ -710,7 +779,7 @@ public class AmqpConsumer extends AmqpAbstractResource<JmsConsumerInfo, Receiver
         }
 
         @Override
-        public void onFailure(Throwable cause) {
+        public void onFailure(ProviderException cause) {
             sheduledTask.cancel(false);
             origRequest.onFailure(cause);
         }
