@@ -19,16 +19,19 @@ package org.apache.qpid.jms.transports.netty;
 import java.io.IOException;
 import java.net.URI;
 import java.security.Principal;
+import java.util.Objects;
 import java.util.concurrent.CountDownLatch;
+import java.util.concurrent.ScheduledExecutorService;
+import java.util.concurrent.ThreadFactory;
 import java.util.concurrent.TimeUnit;
 import java.util.concurrent.atomic.AtomicBoolean;
+import java.util.function.Supplier;
 
 import javax.net.ssl.SSLContext;
 
 import org.apache.qpid.jms.transports.Transport;
 import org.apache.qpid.jms.transports.TransportListener;
 import org.apache.qpid.jms.transports.TransportOptions;
-import org.apache.qpid.jms.transports.TransportSslOptions;
 import org.apache.qpid.jms.transports.TransportSupport;
 import org.apache.qpid.jms.util.IOExceptionSupport;
 import org.slf4j.Logger;
@@ -47,16 +50,12 @@ import io.netty.channel.ChannelPipeline;
 import io.netty.channel.EventLoopGroup;
 import io.netty.channel.FixedRecvByteBufAllocator;
 import io.netty.channel.SimpleChannelInboundHandler;
-import io.netty.channel.epoll.Epoll;
-import io.netty.channel.epoll.EpollEventLoopGroup;
-import io.netty.channel.epoll.EpollSocketChannel;
-import io.netty.channel.kqueue.KQueue;
-import io.netty.channel.kqueue.KQueueEventLoopGroup;
-import io.netty.channel.kqueue.KQueueSocketChannel;
 import io.netty.channel.nio.NioEventLoopGroup;
 import io.netty.channel.socket.nio.NioSocketChannel;
 import io.netty.handler.logging.LoggingHandler;
+import io.netty.handler.proxy.ProxyHandler;
 import io.netty.handler.ssl.SslHandler;
+import io.netty.resolver.NoopAddressResolverGroup;
 import io.netty.util.ReferenceCountUtil;
 import io.netty.util.concurrent.Future;
 import io.netty.util.concurrent.GenericFutureListener;
@@ -75,8 +74,10 @@ public class NettyTcpTransport implements Transport {
     protected EventLoopGroup group;
     protected Channel channel;
     protected TransportListener listener;
+    protected ThreadFactory ioThreadfactory;
     protected int maxFrameSize = DEFAULT_MAX_FRAME_SIZE;
 
+    private final boolean secure;
     private final TransportOptions options;
     private final URI remote;
     private final AtomicBoolean connected = new AtomicBoolean();
@@ -91,9 +92,11 @@ public class NettyTcpTransport implements Transport {
      *        the URI that defines the remote resource to connect to.
      * @param options
      *        the transport options used to configure the socket connection.
+     * @param secure
+     * 		  should the transport enable an SSL layer.
      */
-    public NettyTcpTransport(URI remoteLocation, TransportOptions options) {
-        this(null, remoteLocation, options);
+    public NettyTcpTransport(URI remoteLocation, TransportOptions options, boolean secure) {
+        this(null, remoteLocation, options, secure);
     }
 
     /**
@@ -105,8 +108,10 @@ public class NettyTcpTransport implements Transport {
      *        the URI that defines the remote resource to connect to.
      * @param options
      *        the transport options used to configure the socket connection.
+     * @param secure
+     * 		  should the transport enable an SSL layer.
      */
-    public NettyTcpTransport(TransportListener listener, URI remoteLocation, TransportOptions options) {
+    public NettyTcpTransport(TransportListener listener, URI remoteLocation, TransportOptions options, boolean secure) {
         if (options == null) {
             throw new IllegalArgumentException("Transport Options cannot be null");
         }
@@ -115,64 +120,64 @@ public class NettyTcpTransport implements Transport {
             throw new IllegalArgumentException("Transport remote location cannot be null");
         }
 
+        this.secure = secure;
         this.options = options;
         this.listener = listener;
         this.remote = remoteLocation;
     }
 
     @Override
-    public void connect(SSLContext sslContextOverride) throws IOException {
+    public ScheduledExecutorService connect(final Runnable initRoutine, SSLContext sslContextOverride) throws IOException {
+        if (closed.get()) {
+            throw new IllegalStateException("Transport has already been closed");
+        }
 
         if (listener == null) {
             throw new IllegalStateException("A transport listener must be set before connection attempts.");
         }
 
-        final SslHandler sslHandler;
-        if (isSecure()) {
-            try {
-                TransportSslOptions sslOptions = getSslOptions();
-                sslOptions.setSslContextOverride(sslContextOverride);
-
-                sslHandler = TransportSupport.createSslHandler(getRemoteLocation(), sslOptions);
-            } catch (Exception ex) {
-                // TODO: can we stop it throwing Exception?
-                throw IOExceptionSupport.create(ex);
-            }
-        } else {
-            sslHandler = null;
-        }
-
-        boolean useKQueue = getTransportOptions().isUseKQueue() && KQueue.isAvailable();
-        boolean useEpoll = getTransportOptions().isUseEpoll() && Epoll.isAvailable();
+        TransportOptions transportOptions = getTransportOptions();
+        boolean useKQueue = KQueueSupport.isAvailable(transportOptions);
+        boolean useEpoll = EpollSupport.isAvailable(transportOptions);
 
         if (useKQueue) {
             LOG.trace("Netty Transport using KQueue mode");
-            group = new KQueueEventLoopGroup(1);
+            group = KQueueSupport.createGroup(1, ioThreadfactory);
         } else if (useEpoll) {
             LOG.trace("Netty Transport using Epoll mode");
-            group = new EpollEventLoopGroup(1);
+            group = EpollSupport.createGroup(1, ioThreadfactory);
         } else {
             LOG.trace("Netty Transport using NIO mode");
-            group = new NioEventLoopGroup(1);
+            group = new NioEventLoopGroup(1, ioThreadfactory);
         }
 
         bootstrap = new Bootstrap();
         bootstrap.group(group);
         if (useKQueue) {
-            bootstrap.channel(KQueueSocketChannel.class);
+            KQueueSupport.createChannel(bootstrap);
         } else if (useEpoll) {
-            bootstrap.channel(EpollSocketChannel.class);
+            EpollSupport.createChannel(bootstrap);
         } else {
             bootstrap.channel(NioSocketChannel.class);
         }
         bootstrap.handler(new ChannelInitializer<Channel>() {
             @Override
             public void initChannel(Channel connectedChannel) throws Exception {
-                configureChannel(connectedChannel, sslHandler);
+                if (initRoutine != null) {
+                    try {
+                        initRoutine.run();
+                    } catch (Throwable initError) {
+                        LOG.warn("Error during initialization of channel from provided initialization routine");
+                        connectionFailed(connectedChannel, IOExceptionSupport.create(initError));
+                        throw initError;
+                    }
+                }
+                configureChannel(connectedChannel);
             }
         });
 
-        configureNetty(bootstrap, getTransportOptions());
+        configureNetty(bootstrap, transportOptions);
+        transportOptions.setSslContextOverride(sslContextOverride);
 
         ChannelFuture future = bootstrap.connect(getRemoteHost(), getRemotePort());
         future.addListener(new ChannelFutureListener() {
@@ -199,27 +204,18 @@ public class NettyTcpTransport implements Transport {
                 channel.close().syncUninterruptibly();
                 channel = null;
             }
-            if (group != null) {
-                Future<?> fut = group.shutdownGracefully(0, SHUTDOWN_TIMEOUT, TimeUnit.MILLISECONDS);
-                if (!fut.awaitUninterruptibly(2 * SHUTDOWN_TIMEOUT)) {
-                    LOG.trace("Channel group shutdown failed to complete in allotted time");
-                }
-                group = null;
-            }
 
             throw failureCause;
         } else {
             // Connected, allow any held async error to fire now and close the transport.
-            channel.eventLoop().execute(new Runnable() {
-
-                @Override
-                public void run() {
-                    if (failureCause != null) {
-                        channel.pipeline().fireExceptionCaught(failureCause);
-                    }
+            channel.eventLoop().execute(() -> {
+                if (failureCause != null) {
+                    channel.pipeline().fireExceptionCaught(failureCause);
                 }
             });
         }
+
+        return group;
     }
 
     @Override
@@ -229,7 +225,7 @@ public class NettyTcpTransport implements Transport {
 
     @Override
     public boolean isSecure() {
-        return options.isSSL();
+        return secure;
     }
 
     @Override
@@ -258,12 +254,24 @@ public class NettyTcpTransport implements Transport {
     }
 
     @Override
-    public void send(ByteBuf output) throws IOException {
+    public void write(ByteBuf output) throws IOException {
         checkConnected(output);
+        LOG.trace("Attempted write of buffer: {}", output);
+        channel.write(output, channel.voidPromise());
+    }
 
-        LOG.trace("Attempted write of: {} bytes", output.readableBytes());
+    @Override
+    public void writeAndFlush(ByteBuf output) throws IOException {
+        checkConnected(output);
+        LOG.trace("Attempted write and flush of buffer: {}", output);
+        channel.writeAndFlush(output, channel.voidPromise());
+    }
 
-        channel.writeAndFlush(output);
+    @Override
+    public void flush() throws IOException {
+        checkConnected();
+        LOG.trace("Attempted flush of pending writes");
+        channel.flush();
     }
 
     @Override
@@ -312,6 +320,20 @@ public class NettyTcpTransport implements Transport {
         return maxFrameSize;
     }
 
+    @Override
+    public ThreadFactory getThreadFactory() {
+        return ioThreadfactory;
+    }
+
+    @Override
+    public void setThreadFactory(ThreadFactory factory) {
+        if (isConnected() || channel != null) {
+            throw new IllegalStateException("Cannot set IO ThreadFactory after Transport connect");
+        }
+
+        this.ioThreadfactory = factory;
+    }
+
     //----- Internal implementation details, can be overridden as needed -----//
 
     protected String getRemoteHost() {
@@ -322,7 +344,7 @@ public class NettyTcpTransport implements Transport {
         if (remote.getPort() != -1) {
             return remote.getPort();
         } else {
-            return isSecure() ? getSslOptions().getDefaultSslPort() : getTransportOptions().getDefaultTcpPort();
+            return isSecure() ? getTransportOptions().getDefaultSslPort() : getTransportOptions().getDefaultTcpPort();
         }
     }
 
@@ -345,18 +367,39 @@ public class NettyTcpTransport implements Transport {
         LOG.trace("Channel has gone inactive! Channel is {}", channel);
         if (connected.compareAndSet(true, false) && !closed.get()) {
             LOG.trace("Firing onTransportClosed listener");
-            listener.onTransportClosed();
+            if (channel.eventLoop().inEventLoop()) {
+                listener.onTransportClosed();
+            } else {
+                channel.eventLoop().execute(() -> {
+                    listener.onTransportClosed();
+                });
+            }
+        } else if (!closed.get()) {
+            if (failureCause == null) {
+                failureCause = new IOException("Connection failed");
+            }
+            connectionFailed(channel, failureCause);
         }
     }
 
-    protected void handleException(Channel channel, Throwable cause) throws Exception {
+    protected void handleException(Channel channel, Throwable cause) {
         LOG.trace("Exception on channel! Channel is {}", channel);
         if (connected.compareAndSet(true, false) && !closed.get()) {
             LOG.trace("Firing onTransportError listener");
-            if (failureCause != null) {
-                listener.onTransportError(failureCause);
+            if (channel.eventLoop().inEventLoop()) {
+                if (failureCause != null) {
+                    listener.onTransportError(failureCause);
+                } else {
+                    listener.onTransportError(cause);
+                }
             } else {
-                listener.onTransportError(cause);
+                channel.eventLoop().execute(() -> {
+                    if (failureCause != null) {
+                        listener.onTransportError(failureCause);
+                    } else {
+                        listener.onTransportError(cause);
+                    }
+                });
             }
         } else {
             // Hold the first failure for later dispatch if connect succeeds.
@@ -373,13 +416,13 @@ public class NettyTcpTransport implements Transport {
     //----- State change handlers and checks ---------------------------------//
 
     protected final void checkConnected() throws IOException {
-        if (!connected.get()) {
+        if (!connected.get() || !channel.isActive()) {
             throw new IOException("Cannot send to a non-connected transport.");
         }
     }
 
     private void checkConnected(ByteBuf output) throws IOException {
-        if (!connected.get()) {
+        if (!connected.get() || !channel.isActive()) {
             ReferenceCountUtil.release(output);
             throw new IOException("Cannot send to a non-connected transport.");
         }
@@ -404,10 +447,6 @@ public class NettyTcpTransport implements Transport {
         connectLatch.countDown();
     }
 
-    private TransportSslOptions getSslOptions() {
-        return (TransportSslOptions) getTransportOptions();
-    }
-
     private void configureNetty(Bootstrap bootstrap, TransportOptions options) {
         bootstrap.option(ChannelOption.TCP_NODELAY, options.isTcpNoDelay());
         bootstrap.option(ChannelOption.CONNECT_TIMEOUT_MILLIS, options.getConnectTimeout());
@@ -426,11 +465,38 @@ public class NettyTcpTransport implements Transport {
         if (options.getTrafficClass() != -1) {
             bootstrap.option(ChannelOption.IP_TOS, options.getTrafficClass());
         }
+
+        if (options.getLocalAddress() != null || options.getLocalPort() != 0) {
+            if(options.getLocalAddress() != null) {
+                bootstrap.localAddress(options.getLocalAddress(), options.getLocalPort());
+            } else {
+                bootstrap.localAddress(options.getLocalPort());
+            }
+        }
+        if (options.getProxyHandlerSupplier() != null) {
+            // in case we have a proxy we do not want to resolve the address by ourselves but leave this to the proxy
+            bootstrap.resolver(NoopAddressResolverGroup.INSTANCE);
+        }
     }
 
-    private void configureChannel(final Channel channel, final SslHandler sslHandler) throws Exception {
+    private void configureChannel(final Channel channel) throws Exception {
+        if (options.getProxyHandlerSupplier() != null) {
+            Supplier<ProxyHandler> proxyHandlerSupplier = options.getProxyHandlerSupplier();
+
+            ProxyHandler proxyHandler = proxyHandlerSupplier.get();
+            Objects.requireNonNull(proxyHandler, "No proxy handler was returned by the supplier");
+
+            channel.pipeline().addFirst(proxyHandler);
+        }
         if (isSecure()) {
-            channel.pipeline().addLast(sslHandler);
+            final SslHandler sslHandler;
+            try {
+                sslHandler = TransportSupport.createSslHandler(channel.alloc(), getRemoteLocation(), getTransportOptions());
+            } catch (Exception ex) {
+                throw IOExceptionSupport.create(ex);
+            }
+
+            channel.pipeline().addLast("ssl", sslHandler);
         }
 
         if (getTransportOptions().isTraceBytes()) {
@@ -491,8 +557,15 @@ public class NettyTcpTransport implements Transport {
 
         @Override
         protected void channelRead0(ChannelHandlerContext ctx, ByteBuf buffer) throws Exception {
-            LOG.trace("New data read: {} bytes incoming: {}", buffer.readableBytes(), buffer);
-            listener.onData(buffer);
+            LOG.trace("New data read: {} bytes incomsing: {}", buffer.readableBytes(), buffer);
+            // Avoid all doubts to the contrary
+            if (channel.eventLoop().inEventLoop()) {
+                listener.onData(buffer);
+            } else {
+                channel.eventLoop().execute(() -> {
+                    listener.onData(buffer);
+                });
+            }
         }
     }
 }
