@@ -43,6 +43,7 @@ import java.net.URI;
 import java.util.HashMap;
 import java.util.Map;
 import java.util.Random;
+import java.util.UUID;
 import java.util.concurrent.CountDownLatch;
 import java.util.concurrent.TimeUnit;
 import java.util.concurrent.atomic.AtomicBoolean;
@@ -62,6 +63,7 @@ import javax.jms.ResourceAllocationException;
 import javax.jms.Session;
 
 import org.apache.qpid.jms.JmsConnection;
+import org.apache.qpid.jms.JmsConnectionExtensions;
 import org.apache.qpid.jms.JmsConnectionFactory;
 import org.apache.qpid.jms.JmsConnectionRemotelyClosedException;
 import org.apache.qpid.jms.JmsDefaultConnectionListener;
@@ -75,6 +77,8 @@ import org.apache.qpid.jms.test.testpeer.basictypes.ConnectionError;
 import org.apache.qpid.jms.test.testpeer.matchers.CoordinatorMatcher;
 import org.apache.qpid.jms.test.testpeer.matchers.sections.TransferPayloadCompositeMatcher;
 import org.apache.qpid.jms.util.MetaDataSupport;
+import org.apache.qpid.jms.util.QpidJMSTestRunner;
+import org.apache.qpid.jms.util.Repeat;
 import org.apache.qpid.proton.amqp.Binary;
 import org.apache.qpid.proton.amqp.Symbol;
 import org.apache.qpid.proton.amqp.UnsignedInteger;
@@ -83,7 +87,9 @@ import org.apache.qpid.proton.engine.impl.AmqpHeader;
 import org.hamcrest.Matcher;
 import org.junit.Ignore;
 import org.junit.Test;
+import org.junit.runner.RunWith;
 
+@RunWith(QpidJMSTestRunner.class)
 public class ConnectionIntegrationTest extends QpidJmsTestCase {
     private final IntegrationTestFixture testFixture = new IntegrationTestFixture();
 
@@ -329,6 +335,39 @@ public class ConnectionIntegrationTest extends QpidJmsTestCase {
         }
     }
 
+    @Repeat(repetitions = 1)
+    @Test(timeout = 20000)
+    public void testRemotelyDropConnectionDuringSessionCreationTransacted() throws Exception {
+        try (TestAmqpPeer testPeer = new TestAmqpPeer();) {
+            testPeer.expectSaslAnonymous();
+            testPeer.expectOpen();
+            testPeer.expectBegin();
+
+            ConnectionFactory factory = new JmsConnectionFactory("amqp://localhost:" + testPeer.getServerPort() + "?jms.clientID=foo");
+            Connection connection = factory.createConnection();
+
+            CountDownLatch exceptionListenerFired = new CountDownLatch(1);
+            connection.setExceptionListener(ex -> exceptionListenerFired.countDown());
+
+            // Expect the begin, then drop connection without without a close frame before the tx-coordinator setup.
+            testPeer.expectBegin();
+            testPeer.dropAfterLastHandler();
+
+            try {
+                connection.createSession(true, Session.SESSION_TRANSACTED);
+                fail("Expected exception to be thrown");
+            } catch (JMSException jmse) {
+                // Expected
+            }
+
+            assertTrue("Exception listener did not fire", exceptionListenerFired.await(5, TimeUnit.SECONDS));
+
+            testPeer.waitForAllHandlersToComplete(3000);
+
+            connection.close();
+        }
+    }
+
     @Test(timeout = 20000)
     public void testConnectionPropertiesContainExpectedMetaData() throws Exception {
         try (TestAmqpPeer testPeer = new TestAmqpPeer();) {
@@ -870,6 +909,155 @@ public class ConnectionIntegrationTest extends QpidJmsTestCase {
         }
     }
 
+    @Test(timeout = 20000)
+    public void testConnectionPropertiesExtensionAddedValues() throws Exception {
+        try (TestAmqpPeer testPeer = new TestAmqpPeer();) {
+            final String property1 = "property1";
+            final String property2 = "property2";
+
+            final String value1 = UUID.randomUUID().toString();
+            final String value2 = UUID.randomUUID().toString();
+
+            Matcher<?> connPropsMatcher = allOf(
+                    hasEntry(Symbol.valueOf(property1), value1),
+                    hasEntry(Symbol.valueOf(property2), value2),
+                    hasEntry(AmqpSupport.PRODUCT, MetaDataSupport.PROVIDER_NAME),
+                    hasEntry(AmqpSupport.VERSION, MetaDataSupport.PROVIDER_VERSION),
+                    hasEntry(AmqpSupport.PLATFORM, MetaDataSupport.PLATFORM_DETAILS));
+
+            testPeer.expectSaslAnonymous();
+            testPeer.expectOpen(connPropsMatcher, null, false);
+            testPeer.expectBegin();
+
+            final URI remoteURI = new URI("amqp://localhost:" + testPeer.getServerPort());
+
+            JmsConnectionFactory factory = new JmsConnectionFactory(remoteURI);
+
+            factory.setExtension(JmsConnectionExtensions.AMQP_OPEN_PROPERTIES.toString(), (connection, uri) -> {
+                Map<String, Object> properties = new HashMap<>();
+
+                properties.put(property1, value1);
+                properties.put(property2, value2);
+
+                return properties;
+            });
+
+            Connection connection = factory.createConnection();
+            connection.start();
+
+            testPeer.waitForAllHandlersToComplete(1000);
+            assertNull(testPeer.getThrowable());
+
+            testPeer.expectClose();
+            connection.close();
+        }
+    }
+
+    @Test(timeout = 20000)
+    public void testConnectionPropertiesExtensionAddedValuesOfNonString() throws Exception {
+        try (TestAmqpPeer testPeer = new TestAmqpPeer();) {
+            final String property1 = "property1";
+            final String property2 = "property2";
+
+            final UUID value1 = UUID.randomUUID();
+            final UUID value2 = UUID.randomUUID();
+
+            Matcher<?> connPropsMatcher = allOf(
+                    hasEntry(Symbol.valueOf(property1), value1),
+                    hasEntry(Symbol.valueOf(property2), value2));
+
+            testPeer.expectSaslAnonymous();
+            testPeer.expectOpen(connPropsMatcher, null, false);
+            testPeer.expectBegin();
+
+            final URI remoteURI = new URI("amqp://localhost:" + testPeer.getServerPort());
+
+            JmsConnectionFactory factory = new JmsConnectionFactory(remoteURI);
+
+            factory.setExtension(JmsConnectionExtensions.AMQP_OPEN_PROPERTIES.toString(), (connection, uri) -> {
+                Map<String, Object> properties = new HashMap<>();
+
+                properties.put(property1, value1);
+                properties.put(property2, value2);
+
+                return properties;
+            });
+
+            Connection connection = factory.createConnection();
+            connection.start();
+
+            testPeer.waitForAllHandlersToComplete(1000);
+            testPeer.expectClose();
+            connection.close();
+        }
+    }
+
+    @Test(timeout = 20000)
+    public void testConnectionPropertiesExtensionProtectsClientProperties() throws Exception {
+        try (TestAmqpPeer testPeer = new TestAmqpPeer();) {
+
+            Matcher<?> connPropsMatcher = allOf(
+                    hasEntry(AmqpSupport.PRODUCT, MetaDataSupport.PROVIDER_NAME),
+                    hasEntry(AmqpSupport.VERSION, MetaDataSupport.PROVIDER_VERSION),
+                    hasEntry(AmqpSupport.PLATFORM, MetaDataSupport.PLATFORM_DETAILS));
+
+            testPeer.expectSaslAnonymous();
+            testPeer.expectOpen(connPropsMatcher, null, false);
+            testPeer.expectBegin();
+
+            final URI remoteURI = new URI("amqp://localhost:" + testPeer.getServerPort());
+
+            JmsConnectionFactory factory = new JmsConnectionFactory(remoteURI);
+
+            factory.setExtension(JmsConnectionExtensions.AMQP_OPEN_PROPERTIES.toString(), (connection, uri) -> {
+                Map<String, Object> properties = new HashMap<>();
+
+                properties.put(AmqpSupport.PRODUCT.toString(), "Super-Duper-Qpid-JMS");
+                properties.put(AmqpSupport.VERSION.toString(), "5.0.32.Final");
+                properties.put(AmqpSupport.PLATFORM.toString(), "Commodore 64");
+
+                return properties;
+            });
+
+            Connection connection = factory.createConnection();
+            connection.start();
+
+            testPeer.waitForAllHandlersToComplete(1000);
+            testPeer.expectClose();
+            connection.close();
+        }
+    }
+
+    @Test(timeout = 20000)
+    public void testConnectionFailsWhenUserSuppliesIllegalProperties() throws Exception {
+        try (TestAmqpPeer testPeer = new TestAmqpPeer();) {
+
+            testPeer.expectSaslAnonymous();
+
+            final URI remoteURI = new URI("amqp://localhost:" + testPeer.getServerPort());
+
+            JmsConnectionFactory factory = new JmsConnectionFactory(remoteURI);
+
+            factory.setExtension(JmsConnectionExtensions.AMQP_OPEN_PROPERTIES.toString(), (connection, uri) -> {
+                Map<String, Object> properties = new HashMap<>();
+
+                properties.put("not-amqp-encodable", factory);
+
+                return properties;
+            });
+
+            Connection connection = factory.createConnection();
+
+            try {
+                connection.start();
+                fail("Should not be able to connect when illegal types are in the properties");
+            } catch (JMSException ex) {
+            }
+
+            testPeer.waitForAllHandlersToComplete(1000);
+        }
+    }
+
     @Ignore("Disabled due to requirement of hard coded port")
     @Test(timeout = 20000)
     public void testLocalPortOption() throws Exception {
